package ssod;

import java.io.DataInputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.net.ServerSocket;
import java.net.Socket;
import java.security.InvalidKeyException;
import java.security.Key;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.security.spec.InvalidKeySpecException;
import java.security.spec.KeySpec;
import java.util.Arrays;
import java.util.Properties;

import javax.crypto.BadPaddingException;
import javax.crypto.Cipher;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.NoSuchPaddingException;
import javax.crypto.SecretKeyFactory;
import javax.crypto.spec.DESedeKeySpec;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Sample code that listens to notifications sent by Windows Services for Unix,
 * decrypts them and displays the information (username and password).
 * 
 * It expects that file ssod.properties exists and contains the following information:
 * 
 * port=6677
 * secret=...
 * 
 * (Secret should match the value set on the Windows server using either
 * Microsoft Identity Management for UNIX console or "psadmin" command.)
 * 
 * EXPERIMENTAL/PROTOTYPE CODE. No warranties. Use at your own risk.
 * 
 * @author pavol.mederly@gmail.com
 */

public class Server {
	
	private static Logger logger = LoggerFactory.getLogger(Server.class);

	public static void main(String[] args) throws Exception {

		Properties prop = new Properties();
		prop.load(new FileReader("ssod.properties"));
		
		int port = Integer.parseInt(prop.getProperty("port", "6677"));
		String secret = prop.getProperty("secret");
		
		ServerSocket sock = new ServerSocket(port);
		logger.info("Listening on port " + port);
		
		for (;;) {
			Socket client = sock.accept();
			logger.info("Got client connection from " + client.getRemoteSocketAddress());
		
			InputStream is = client.getInputStream();
			OutputStream os = client.getOutputStream();
			
			byte[] random = generateRandomValue();
			os.write(random);
			int versionNumber = readInt(is);
			int messageSize = readInt(is);
			logger.info("versionNumber = " + versionNumber + ", messageSize = " + messageSize);
			
			byte[] data = readBlock(is, messageSize - 8);
			int packetType = parseInt(data);
			logger.info("packetType = " + packetType);		
			
			byte[] encryptedData = new byte[data.length - 4];
			System.arraycopy(data, 4, encryptedData, 0, encryptedData.length);
			parseRequest(encryptedData, random, secret);
			
//			logger.info("Waiting...");
//			Thread.sleep(5000);
			logger.info("Done, sending response.");

			os.write(prepareResponse(0));
			client.close();
		}
	}

	private static byte[] prepareResponse(int error) {
		byte[] retval = new byte[16];
		System.arraycopy(createInt(0), 0, retval, 0, 4);	//version = 0
		System.arraycopy(createInt(16), 0, retval, 4, 4);	//size = 16 bytes
		System.arraycopy(createInt(1), 0, retval, 8, 4);	//type = 1 (response)
		System.arraycopy(createInt(error), 0, retval, 12, 4);		
		return retval;
	}

	private static byte[] createInt(int value) {
		byte[] retval = new byte[4];
		retval[0] = (byte) ((value>>>24) & 0xff);
		retval[1] = (byte) ((value>>>16) & 0xff);
		retval[2] = (byte) ((value>>>8) & 0xff);
		retval[3] = (byte) (value & 0xff);
		return retval;
	}

	private static int parseInt(byte[] data) {
		int value = 0;
		for (int i = 0; i < 4; i++) {
		   value = (value << 8) | (data[i] & 0xff);
		}
		logger.info("- parsing " + dump(data, 4) + " as " + value);
		return value;
	}

	private static byte[] readBlock(InputStream is, int length) throws IOException {
		DataInputStream dis = new DataInputStream(is);
		byte[] data = new byte[length];
		dis.readFully(data);
		return data;
	}

	private static int readInt(InputStream is) throws IOException {
		return parseInt(readBlock(is, 4));
	}

	private static byte[] generateRandomValue() {
		final int LENGTH = 8;
		byte[] retval = new byte[LENGTH];
		for (int i = 0; i < LENGTH; i++) {
			retval[i] = (byte) (Math.random() * 256 - 128);
		}
		logger.info(" - random value: " + dump(retval));
		return retval;
	}

	private static String dump(byte[] data) {
		return dump(data, data.length);
	}

	private static String dump(byte[] data, int length) {
		StringBuilder sb = new StringBuilder();
		for (int i = 0; i < length; i++) {
			if (i > 0) {
				sb.append(", ");
			}
			sb.append(data[i]);
		}
		return sb.toString();
	}
	
	private static void parseRequest(byte[] encryptedData, byte[] random, String secret) throws NoSuchAlgorithmException, UnsupportedEncodingException, InvalidKeyException, NoSuchPaddingException, InvalidKeySpecException, IllegalBlockSizeException, BadPaddingException {
		
//    	byte[] retval = new byte[encryptedData.length];
    	
        byte[] keyIn = makeHash(random, encryptedData, secret);
        keyIn = extendHashForKey(keyIn);

        KeySpec ks = new DESedeKeySpec(keyIn);
        SecretKeyFactory skf = SecretKeyFactory.getInstance("DESede");
        Key key = skf.generateSecret(ks);
        
        //Cipher cipher = Cipher.getInstance("DESede/NONE/NoPadding");
        Cipher cipher = Cipher.getInstance("DESede/ECB/NoPadding");
        logger.info("- alg = " + cipher.getAlgorithm());
        logger.info(" = alg params = " + cipher.getParameters());
        
        byte[] desEncryptedData = new byte[encryptedData.length-8];
        System.arraycopy(encryptedData, 8, desEncryptedData, 0, desEncryptedData.length);

        logger.info("encryptedData length = " + encryptedData.length);
        logger.info("desEncryptedData length = " + desEncryptedData.length);
        logger.info("- content = " + dump(desEncryptedData));
        
        cipher.init(Cipher.DECRYPT_MODE, key);
        byte[] plainText = cipher.doFinal(desEncryptedData);
        
        int usernameEnd = find(plainText, 0, 0);
        int passwordEnd = find(plainText, usernameEnd+1, 0);
        if (usernameEnd < 0 || passwordEnd < 0) {
        	logger.error("Invalid format; usernameEnd = " + usernameEnd + ", passwordEnd = " + passwordEnd);
        	return;
        }
        byte[] username = new byte[usernameEnd];
        System.arraycopy(plainText, 0, username, 0, username.length);
        byte[] password = new byte[passwordEnd-usernameEnd-1];
        System.arraycopy(plainText, usernameEnd+1, password, 0, password.length);
        
        String usernameAsString = new String(username, "UTF-8");
        String passwordAsString = new String(password, "UTF-8");
        
        logger.info("UsernameEnd = " + usernameEnd + ", passwordEnd = " + passwordEnd);
        logger.info("Plaintext size = " + plainText.length + ", data = " + plainText);
        logger.info("Username = " + usernameAsString + ", password = " + passwordAsString);


	    byte[] sigReceived = new byte[20];
	    System.arraycopy(plainText, passwordEnd+1, sigReceived, 0, 20);
	    logger.info("SigReceived size = " + sigReceived.length + ", data = " + dump(sigReceived));

	    MessageDigest md = MessageDigest.getInstance("SHA-1");
	    int[] keyTable = DesKeyUtil.generate3DesKeyTable(keyIn);
        md.update(intArrayToByteArray(keyTable));
	    md.update(createInt(0));			// version
	    md.update(createInt(encryptedData.length + 12));	// msg size
	    md.update(createInt(0));			// msg type
	    md.update(username);
	    md.update(password);
	    
	    check(md, sigReceived);	    

	}
	
	private static byte[] intArrayToByteArray(int[] data) {
		byte[] retval = new byte[data.length * 4];
		for (int i = 0; i < data.length; i++) {
			retval[4*i] = (byte) (data[i]>>>24 & 0xff);
			retval[4*i+1] = (byte) (data[i]>>>16 & 0xff);
			retval[4*i+2] = (byte) (data[i]>>>8 & 0xff);
			retval[4*i+3] = (byte) (data[i] & 0xff);
		}
		
		logger.info("KeyTable size = " + retval.length);
		return retval;
	}

	private static void check(MessageDigest md, byte[] sigReceived) {
	    byte[] sigExpected = md.digest();
	    
	    logger.info("SigExpected size = " + sigExpected.length + ", data = " + dump(sigExpected));
	    boolean ok = Arrays.equals(sigReceived, sigExpected);
	    logger.info("Match = " + ok);
	}

	private static byte[] revertWords(byte[] original) {
		byte[] retval = new byte[original.length];
		for (int i = 0; i < retval.length; i+=4) {
			retval[i] = original[i+3];
			retval[i+1] = original[i+2];
			retval[i+2] = original[i+1];
			retval[i+3] = original[i];
		}
		return retval;
	}

	private static int find(byte[] data, int from, int what) {
		while (from < data.length) {
			if (data[from] == what) {
				return from;
			}
			from++;
		}
		return -1;
	}

	private static byte[] extendHashForKey(byte[] keyIn) throws NoSuchAlgorithmException {

		MessageDigest md1 = MessageDigest.getInstance("SHA-1");
		MessageDigest md2 = MessageDigest.getInstance("SHA-1");

		byte[] buf1 = new byte[64];
		byte[] buf2 = new byte[64];
		
		for (int i = 0; i < 64; i++) {
			buf1[i] = 0x36;
			buf2[i] = 0x5c;
			if (i < 20) {
				buf1[i] ^= keyIn[i];
				buf2[i] ^= keyIn[i];
			}
		}
		md1.update(buf1);
		md2.update(buf2);

		byte[] retval = new byte[24];
		System.arraycopy(md1.digest(), 0, retval, 0, 20);
		System.arraycopy(md2.digest(), 0, retval, 20, 4);
		return retval;
	}

	static byte[] makeHash(byte[] random, byte[] encryptedData, String secret) throws NoSuchAlgorithmException, UnsupportedEncodingException {
		MessageDigest md = MessageDigest.getInstance("SHA-1");
		
	    md.update(random);
	    md.update(secret.getBytes("ASCII"));
	    md.update(encryptedData, 0, 8);
	    md.update(new byte[] { 0, 0, 0, 0 });
	    md.update(new byte[] { 0, 0, 0, 0 });
	    return md.digest();
	} 
}


